/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid;

import static java.util.Objects.*;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.stereotype.Component;

import de.ozgcloud.document.bescheid.Bescheid.SendBy;
import de.ozgcloud.document.bescheid.binaryfile.BinaryFileService;
import de.ozgcloud.document.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.document.bescheid.common.user.UserProfileService;
import de.ozgcloud.document.bescheid.vorgang.VorgangId;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.BescheidDocumentCreatedEvent;
import de.ozgcloud.document.Document;
import de.ozgcloud.document.DocumentService;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Component
@RequiredArgsConstructor
class BescheidEventListener {

	public static final String CREATE_BESCHEID_ORDER = "CREATE_BESCHEID";
	public static final String DELETE_BESCHEID_ORDER = "DELETE_BESCHEID";
	public static final String UPDATE_BESCHEID_ORDER = "UPDATE_BESCHEID";
	public static final String CREATE_BESCHEID_DOCUMENT_ORDER = "CREATE_BESCHEID_DOCUMENT";
	public static final String SEND_BESCHEID_ORDER = "SEND_BESCHEID";

	public static final Predicate<Command> IS_CREATE_BESCHEID_COMMAND = command -> CREATE_BESCHEID_ORDER.equals(command.getOrder());
	private static final String IS_CREATE_BESCHEID = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_CREATE_BESCHEID_COMMAND.test(event.getSource())}";

	public static final Predicate<Command> IS_DELETE_BESCHEID_COMMAND = command -> DELETE_BESCHEID_ORDER.equals(command.getOrder());
	private static final String IS_DELETE_BESCHEID = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_DELETE_BESCHEID_COMMAND.test(event.getSource())}";

	public static final Predicate<Command> IS_UPDATE_BESCHEID_COMMAND = command -> UPDATE_BESCHEID_ORDER.equals(command.getOrder());
	private static final String IS_UPDATE_BESCHEID = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_UPDATE_BESCHEID_COMMAND.test(event.getSource())}";

	public static final Predicate<Command> IS_CREATE_BESCHEID_DOCUMENT_COMMAND = command -> CREATE_BESCHEID_DOCUMENT_ORDER.equals(command.getOrder());
	private static final String IS_CREATE_BESCHEID_DOCUMENT = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_CREATE_BESCHEID_DOCUMENT_COMMAND.test(event.getSource())}";

	public static final Predicate<Command> IS_SEND_BESCHEID_COMMAND = command -> nonNull(command) && SEND_BESCHEID_ORDER.equals(command.getOrder());
	private static final String IS_SEND_BESCHEID = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_SEND_BESCHEID_COMMAND.test(event.getSource())}";
	private static final String IS_SEND_BESCHEID_ORDER = "{T(de.ozgcloud.document.bescheid.BescheidEventListener).IS_SEND_BESCHEID_COMMAND.test(event.getCommand())}";

	static final String VORGANG_ID_BODY_KEY = "vorgangId";
	static final String BESCHEID_VOM_BODY_KEY = "bescheidVom";
	static final String GENEHMIGT_BODY_KEY = "genehmigt";

	private static final String LOG_MESSAGE_TEMPLATE = "{}. Command failed.";
	private static final String ERROR_MESSAGE_TEMPLATE = "Error on executing %s Command.";

	private final BescheidService service;
	private final BinaryFileService fileService;
	private final AttachedItemService attachedItemService;
	private final DocumentService documentService;

	private final ApplicationEventPublisher eventPublisher;
	private final CurrentUserService userService;
	private final UserProfileService userProfileService;

	@EventListener(condition = IS_CREATE_BESCHEID)
	public void onCreateBescheidCommand(CommandCreatedEvent event) {
		runWithSecurityContext(event.getSource(), this::doCreateBescheid);
	}

	void doCreateBescheid(Command command) {
		var createdItemId = service.createBescheidDraft(command);
		eventPublisher.publishEvent(new BescheidCreatedEvent(command, createdItemId));
	}

	@EventListener(condition = IS_DELETE_BESCHEID)
	public void onDeleteBescheid(CommandCreatedEvent event) {
		runWithSecurityContext(event.getSource(), this::doDeleteBescheid);
	}

	void doDeleteBescheid(Command command) {
		attachedItemService.deleteBescheidDraft(command);
		eventPublisher.publishEvent(new BescheidDeletedEvent(command));
	}

	@EventListener(condition = IS_UPDATE_BESCHEID)
	public void onUpdateBescheidCommand(CommandCreatedEvent event) {
		runWithSecurityContext(event.getSource(), this::doUpdateBescheid);
	}

	void doUpdateBescheid(Command command) {
		attachedItemService.updateBescheidDraft(command, getDocument(command));
		eventPublisher.publishEvent(new BescheidUpdatedEvent(command));
	}

	Optional<Document> getDocument(Command command) {
		return Optional.ofNullable(MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_BESCHEID_DOCUMENT))
				.map(StringUtils::trimToNull)
				.map(documentService::getDocument);
	}

	@EventListener(condition = IS_CREATE_BESCHEID_DOCUMENT)
	public void onCreatedBescheidDocument(CommandCreatedEvent event) {
		runWithSecurityContext(event.getSource(), this::doCreateBescheidDocument);
	}

	void doCreateBescheidDocument(Command command) {
		var bescheid = doCreateBescheidBiz(command);
		var bescheidDocument = documentService.createBescheidDocument(command, bescheid);
		eventPublisher.publishEvent(new BescheidDocumentCreatedEvent(command, bescheidDocument));
	}

	BescheidResponse doCreateBescheidBiz(@NonNull Command command) {
		var bescheid = service.createBescheid(createRequest(command));
		return fileService.uploadBescheidFile(bescheid);
	}

	@EventListener(condition = IS_SEND_BESCHEID)
	public void onSendBescheidCommand(CommandCreatedEvent event) {
		runWithSecurityContext(event.getSource(), this::doSendBescheid);
	}

	void doSendBescheid(Command command) {
		var bescheidItem = attachedItemService.getItem(command.getRelationId());
		if (SendBy.MANUAL.hasValue(bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY))) {
			service.sendBescheidManually(bescheidItem, command);
		} else if (SendBy.NACHRICHT.hasValue(bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY))) {
			service.sendBescheidPostfachMail(bescheidItem, command);
		} else {
			throw new TechnicalException("Bescheid has unexpected sendBy value: '%s'. Allowed values are %s."
					.formatted(bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY),
							Arrays.stream(SendBy.values()).map(SendBy::name).collect(Collectors.joining(","))));
		}
		eventPublisher.publishEvent(new BescheidSentEvent(command));
	}

	BescheidRequest createRequest(Command command) {
		var eventBody = command.getBodyObject();
		var builder = BescheidRequest.builder();

		builder.vorgangId(VorgangId.from(command.getVorgangId()));
		Optional.ofNullable(MapUtils.getString(eventBody, BESCHEID_VOM_BODY_KEY))
				.map(LocalDate::parse).ifPresent(builder::bescheidVom);
		Optional.ofNullable(MapUtils.getBoolean(eventBody, GENEHMIGT_BODY_KEY))
				.ifPresentOrElse(builder::genehmigt, () -> builder.genehmigt(true));
		builder.createFor(userProfileService.getUserProfile());

		return builder.build();
	}

	@EventListener(condition = IS_SEND_BESCHEID_ORDER)
	public void onBescheidSent(CommandExecutedEvent event) {
		runWithSecurityContext(event.getCommand(), this::setAntragBewilligung);
	}

	void setAntragBewilligung(Command command) {
		service.setAntragBewilligung(attachedItemService.getItem(command.getRelationId()));
	}

	void runWithSecurityContext(Command command, Consumer<Command> commandExecutor) {
		SecurityContext prevContext = null;
		try {
			prevContext = userService.startSecurityContext(command);
			commandExecutor.accept(command);
		} catch (Exception e) {
			var errorMessage = ERROR_MESSAGE_TEMPLATE.formatted(command.getOrder());
			LOG.error(LOG_MESSAGE_TEMPLATE, errorMessage, e);
			eventPublisher.publishEvent(new CommandFailedEvent(command.getId(), buildErrorMessage(errorMessage, e)));
		} finally {
			userService.resetSecurityContext(prevContext);
		}
	}

	private String buildErrorMessage(String message, Exception cause) {
		try {
			StringBuilder sb = new StringBuilder(message);

			if (StringUtils.isNotBlank(cause.getMessage())) {
				sb.append(" Cause: ").append(cause.getMessage());
			}

			return sb.toString();
		} catch (Exception e2) {
			LOG.error("Error in building Error Message (sick).", e2);
			return message;
		}
	}

}
