/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid.attributes;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.document.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.grpc.clientAttribute.ClientAttributeServiceGrpc.ClientAttributeServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcAccessPermission;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttribute;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcSetClientAttributeRequest;

class ClientAttributeRemoteServiceTest {

	private static final String VORGANG_ID = "vorgangId";
	private static final String ATTRIBUTE_NAME = "attributeName";

	@Spy
	@InjectMocks
	private ClientAttributeRemoteService service;

	@Mock
	private ClientAttributeServiceBlockingStub serviceBlockingStub;
	@Mock
	private BescheidCallContextAttachingInterceptor bescheidCallContextInterceptor;

	@Nested
	class TestSetBooleanReadOnlyClientAttribute {

		@Captor
		private ArgumentCaptor<BescheidCallContextAttachingInterceptor> interceptorCaptor;

		@BeforeEach
		void setUp() {
			when(serviceBlockingStub.withInterceptors(any())).thenReturn(serviceBlockingStub);
		}

		@Test
		void shouldSetInterceptors() {
			service.setBooleanReadOnlyClientAttribute(VORGANG_ID, ATTRIBUTE_NAME, true);

			verify(serviceBlockingStub).withInterceptors(interceptorCaptor.capture());
			assertThat(interceptorCaptor.getValue()).isEqualTo(bescheidCallContextInterceptor);
		}

		@Test
		void shouldCallBuildRequest() {
			service.setBooleanReadOnlyClientAttribute(VORGANG_ID, ATTRIBUTE_NAME, true);

			verify(service).buildRequest(VORGANG_ID, ATTRIBUTE_NAME, true);
		}

		@Test
		void shouldCallSetMethod() {
			var request = GrpcSetClientAttributeRequest.newBuilder().build();
			doReturn(request).when(service).buildRequest(any(), any(), anyBoolean());

			service.setBooleanReadOnlyClientAttribute(VORGANG_ID, ATTRIBUTE_NAME, true);

			verify(serviceBlockingStub).set(request);

		}
	}

	@Nested
	class TestBuildRequest {

		@Mock
		private GrpcClientAttribute clientAttribute;

		@Test
		void shouldSetVorgangId() {
			var result = buildRequest();

			assertThat(result.getVorgangId()).isEqualTo(VORGANG_ID);
		}

		@Test
		void shouldCallBuildClientAttribute() {
			buildRequest();

			verify(service).buildClientAttribute(ATTRIBUTE_NAME, true);
		}

		@Test
		void shouldSetClientAttribute() {
			doReturn(clientAttribute).when(service).buildClientAttribute(any(), anyBoolean());

			var result = buildRequest();

			assertThat(result.getAttribute()).isEqualTo(clientAttribute);
		}

		private GrpcSetClientAttributeRequest buildRequest() {
			return service.buildRequest(VORGANG_ID, ATTRIBUTE_NAME, true);
		}
	}

	@Nested
	class TestBuildClientAttribute {

		@Test
		void shouldSetClientName() {
			var result = service.buildClientAttribute(ATTRIBUTE_NAME, true);

			assertThat(result.getClientName()).isEqualTo(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		@Test
		void shouldSetAccess() {
			var result = service.buildClientAttribute(ATTRIBUTE_NAME, true);

			assertThat(result.getAccess()).isEqualTo(GrpcAccessPermission.READ_ONLY);
		}

		@Test
		void shouldSetAttributeName() {
			var result = service.buildClientAttribute(ATTRIBUTE_NAME, true);

			assertThat(result.getAttributeName()).isEqualTo(ATTRIBUTE_NAME);
		}

		@Test
		void shouldSetValue() {
			var result = service.buildClientAttribute(ATTRIBUTE_NAME, true);

			assertThat(result.getValue().getBoolValue()).isTrue();
		}
	}
}