package de.ozgcloud.document.bescheid.smartdocuments;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.Document;
import org.w3c.dom.Text;

import de.ozgcloud.document.bescheid.BescheidRequestTestFactory;
import de.ozgcloud.document.bescheid.BescheidResponseTestFactory;
import de.ozgcloud.document.bescheid.common.user.UserProfileTestFactory;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsRequest.CustomerData.UserData;
import de.ozgcloud.document.bescheid.vorgang.VorgangTestFactory;
import lombok.SneakyThrows;

class SmartDocumentsBescheidRemoteServiceTest {

	@Spy
	@InjectMocks
	private SmartDocumentsBescheidRemoteService service;

	@Mock
	private SmartDocumentsProperties properties;
	@Mock
	private Text xmlText;
	@Mock
	private File xmlFile;

	@Nested
	class TestBuildBescheid {

		@Test
		void shouldFillBescheid() {
			var bescheid = service.buildBescheid(BescheidRequestTestFactory.create(), SmartDocumentsResponseTestFactory.create());

			assertThat(bescheid).usingRecursiveComparison().ignoringFields("serviceKonto", "nachrichtSubject", "bescheidFileId", "attachments")
					.isEqualTo(BescheidResponseTestFactory.createBuilder().nachrichtText(Optional.empty()).build());
		}
	}

	@Nested
	class TestCreateRequest {

		@Test
		void shouldFillUserData() {
			var request = service.createRequest(BescheidRequestTestFactory.create(), VorgangTestFactory.create());

			assertThat(request.getCustomerData().getUserData()).isNotNull()
					.extracting(UserData::getEmail).isEqualTo(UserProfileTestFactory.EMAIL);
		}
	}

	@Nested
	class TestGetNachrichtText {

		@Test
		void shouldCallGetXMLFile() {
			var response = SmartDocumentsResponseTestFactory.create();

			service.getNachrichtText(response);

			verify(service).getXMLFile(response);
		}

		@Test
		void shouldCallExtractText() {
			doReturn(Optional.of(xmlFile)).when(service).getXMLFile(any());

			service.getNachrichtText(SmartDocumentsResponseTestFactory.create());

			verify(service).extractTextFromXmlFile(xmlFile);
		}
	}

	@Nested
	class TestExtractTextFromXmlFileSuccessfully {

		private static final String NACHRICHT_TEXT = "NachrichtText";

		@Mock
		private Document document;

		@SneakyThrows
		@BeforeEach
		void init() {
			doReturn(Optional.of(document)).when(service).parseXml(any());
		}

		@SneakyThrows
		@Test
		void shouldCallParseXml() {
			service.extractTextFromXmlFile(xmlFile);

			verify(service).parseXml(xmlFile);
		}

		@Test
		void shouldCallGetNachrichtenText() {
			doReturn(Optional.of(xmlText)).when(service).getNachrichtenText(any());

			service.extractTextFromXmlFile(xmlFile);

			verify(service).getNachrichtenText(document);
		}

		@Test
		void shouldReturnNachrichtenText() {
			doReturn(Optional.of(xmlText)).when(service).getNachrichtenText(any());
			when(xmlText.getTextContent()).thenReturn(NACHRICHT_TEXT);

			var text = service.extractTextFromXmlFile(xmlFile);

			assertThat(text).contains(NACHRICHT_TEXT);
		}

		@Test
		void shouldNotCallGetFieldTemplateText() {
			doReturn(Optional.of(xmlText)).when(service).getNachrichtenText(any());

			service.extractTextFromXmlFile(xmlFile);

			verify(service, never()).getFieldTemplateText(document);
		}

		@Test
		void shouldCallGetFieldTemplateText() {
			doReturn(Optional.empty()).when(service).getNachrichtenText(any());
			doReturn(Optional.of(xmlText)).when(service).getFieldTemplateText(any());

			service.extractTextFromXmlFile(xmlFile);

			verify(service).getFieldTemplateText(document);
		}

		@Test
		void shouldReturnFieldTemplateText() {
			doReturn(Optional.empty()).when(service).getNachrichtenText(any());
			doReturn(Optional.of(xmlText)).when(service).getFieldTemplateText(any());
			when(xmlText.getTextContent()).thenReturn(NACHRICHT_TEXT);

			var text = service.extractTextFromXmlFile(xmlFile);

			assertThat(text).contains(NACHRICHT_TEXT);
		}
	}

	@Nested
	class TestExtractTextFormXmlFileFails {

		@Mock
		private Document document;

		@SneakyThrows
		@Test
		void shouldHandleMissingXmlDocument() {
			doReturn(Optional.empty()).when(service).parseXml(any());

			var text = service.extractTextFromXmlFile(xmlFile);

			assertThat(text).isEmpty();
		}

		@Test
		@SneakyThrows
		void shouldExpectMissingTextNode() {
			doReturn(Optional.of(document)).when(service).parseXml(any());
			doReturn(Optional.empty()).when(service).getFieldTemplateText(any());

			var text = service.extractTextFromXmlFile(xmlFile);

			assertThat(text).isEmpty();
		}
	}

	@Nested
	class TestGetNachrichtenText {

		@Mock
		private Text text;
		@Mock
		private Document document;

		@BeforeEach
		void init() {
			doReturn(Optional.of(text)).when(service).evaluateXPath(any(), any());
		}

		@Test
		void shouldCallEvaluatePath() {
			getNachrichtenText();

			verify(service).evaluateXPath(document, SmartDocumentsBescheidRemoteService.NACHRICHTEN_TEXT_EXPRESSION);
		}

		@Test
		void shouldReturnText() {
			var result = getNachrichtenText();

			assertThat(result).contains(text);
		}

		private Optional<Text> getNachrichtenText() {
			return service.getNachrichtenText(document);
		}
	}

	@Nested
	class TestGetFieldTemplateText {

		@Mock
		private Text text;
		@Mock
		private Document document;

		@BeforeEach
		void init() {
			doReturn(Optional.of(text)).when(service).evaluateXPath(any(), any());
		}

		@Test
		void shouldCallEvaluatePath() {
			getFieldTemplateText();

			verify(service).evaluateXPath(document, SmartDocumentsBescheidRemoteService.FIELD_TEMPLATE_EXPRESSION);
		}

		@Test
		void shouldReturnText() {
			var result = getFieldTemplateText();

			assertThat(result).contains(text);
		}

		private Optional<Text> getFieldTemplateText() {
			return service.getFieldTemplateText(document);
		}
	}
}
