/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.common.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.document.bescheid.Bescheid;
import de.ozgcloud.document.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.document.bescheid.BescheidMapper;
import de.ozgcloud.document.bescheid.BescheidTestFactory;
import de.ozgcloud.document.bescheid.vorgang.VorgangId;
import de.ozgcloud.document.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.Document;
import de.ozgcloud.document.DocumentTestFactory;
import io.grpc.StatusRuntimeException;

class AttachedItemServiceTest {

	@Spy
	@InjectMocks
	private AttachedItemService service;
	@Mock
	private AttachedItemMapper mapper;
	@Mock
	private VorgangAttachedItemRemoteService remoteService;

	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;
	@Mock
	private BescheidMapper bescheidMapper;

	@Nested
	class TestFindBescheidItem {

		@Test
		void shouldCallGetBescheidDraft() {
			service.findBescheidItem(VorgangTestFactory.ID);

			verify(remoteService).findBescheidDraft(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnBescheidItem() {
			var expected = AttachedItemTestFactory.createBescheid();
			when(remoteService.findBescheidDraft(any(VorgangId.class))).thenReturn(Optional.of(expected));

			var actual = service.findBescheidItem(VorgangTestFactory.ID);

			assertThat(actual).contains(expected);
		}
	}

	@Nested
	class TestCreateBescheidDraft {

		private static final String CREATED_ATTACHED_ITEM_ID = "attached-item-id";
		private static final Bescheid BESCHEID = BescheidTestFactory.create();

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();

		@Mock
		private AttachedItem updatedBescheidItem;

		@Test
		void shouldCallGetBescheidDraft() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createAttachedItem(any());

			service.createBescheidDraft(BESCHEID);

			verify(remoteService).findBescheidDraft(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallMergeBescheidDrafts() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).overrideAttachedItem(any());
			when(remoteService.findBescheidDraft(any(VorgangId.class))).thenReturn(Optional.of(bescheidItem));

			service.createBescheidDraft(BESCHEID);

			verify(service).mergeBescheidDrafts(bescheidItem, BESCHEID);
		}

		@Test
		void shouldCallOverrideAttachedItem() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).overrideAttachedItem(any());
			when(remoteService.findBescheidDraft(any(VorgangId.class))).thenReturn(Optional.of(bescheidItem));
			doReturn(updatedBescheidItem).when(service).mergeBescheidDrafts(any(), any());

			service.createBescheidDraft(BESCHEID);

			verify(service).overrideAttachedItem(updatedBescheidItem);
		}

		@Test
		void shouldCallCreateBescheidDraft() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createAttachedItem(any());

			service.createBescheidDraft(BESCHEID);

			verify(service).createBescheidDraft(BESCHEID);
		}

	}

	@Nested
	class TestCreateAttachedItem {

		private static final Bescheid BESCHEID = BescheidTestFactory.create();
		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

		@Mock
		private OzgCloudCommand createAttachedItemCommand;
		@Mock
		private OzgCloudCommand finishedOzgCloudCommand;

		@BeforeEach
		void init() {
			when(commandService.createAndWaitUntilDone(any())).thenReturn(finishedOzgCloudCommand);
			when(finishedOzgCloudCommand.getCreatedResource()).thenReturn(AttachedItemTestFactory.ID);
			when(bescheidMapper.mapFromBescheid(any())).thenReturn(BESCHEID_ITEM);
		}

		@Test
		void shouldCallMapFromBescheid() {
			createAttachedItem();

			verify(bescheidMapper).mapFromBescheid(BESCHEID);
		}

		@Test
		void shouldCallBuildCreateAttachedItemCommand() {
			doReturn(createAttachedItemCommand).when(service).buildCreateAttachedItemCommand(any());

			createAttachedItem();

			verify(service).buildCreateAttachedItemCommand(BESCHEID_ITEM);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(createAttachedItemCommand).when(service).buildCreateAttachedItemCommand(any());

			createAttachedItem();

			verify(commandService).createAndWaitUntilDone(createAttachedItemCommand);
		}

		@Test
		void shouldReturnCreatedResource() {
			var result = createAttachedItem();

			assertThat(result).isEqualTo(AttachedItemTestFactory.ID);
		}

		private String createAttachedItem() {
			return service.createAttachedItem(BESCHEID);
		}
	}

	@Nested
	class TestBuildCreateAttachedItemCommand {

		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

		@Test
		void shouldCallVorgangIdMapper() {
			service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangId() {
			var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
			when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

			var result = service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			assertThat(result.getVorgangId()).isEqualTo(expectedVorgangId);
		}

		@Test
		void shouldCallRelationIdMapper() {
			service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			verify(commandMapper).mapRelationId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetRelationId() {
			var expectedRelationId = "relationId";
			when(commandMapper.mapRelationId(any())).thenReturn(GenericId.from(expectedRelationId));

			var result = service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			assertThat(result.getRelationId()).hasToString(expectedRelationId);
		}

		@Test
		void shouldSetOrder() {
			var result = service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			assertThat(result.getOrder()).isEqualTo(AttachedItemService.CREATE_ATTACHED_ITEM_ORDER);
		}

		@Test
		void shouldCallAttachedItemMapper() {
			service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			verify(mapper).toMap(BESCHEID_ITEM);
		}

		@Test
		void shouldSetBodyObject() {
			var itemMap = Map.<String, Object>of("key", "value");
			when(mapper.toMap(any())).thenReturn(itemMap);

			var result = service.buildCreateAttachedItemCommand(BESCHEID_ITEM);

			assertThat(result.getBodyObject()).containsAllEntriesOf(itemMap);
		}
	}

	@Nested
	class TestUpdateBescheidDraft {

		@Mock
		private OzgCloudCommand updateAttachedItemCommand;

		private final Command command = CommandTestFactory.createBuilder().relationId(AttachedItemTestFactory.ID).build();
		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();
		private final Optional<Document> document = Optional.of(DocumentTestFactory.create());

		@BeforeEach
		void init() {
			when(remoteService.getItem(any())).thenReturn(bescheidItem);
		}

		@Test
		void shouldCallGetBescheid() {
			doUpdateBescheidDraft();

			verify(remoteService, times(2)).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallValidateBescheidStatus() {
			doUpdateBescheidDraft();

			verify(service).validateBescheidStatus(bescheidItem);
		}

		@Test
		void shouldCallBuildUpdateBescheidItemMap() {
			doUpdateBescheidDraft();

			verify(service).buildBescheidItemAsMap(command, document);
		}

		@Test
		void shouldCallBuildAttachedItemAsMap() {
			var itemMap = Map.<String, Object>of("key", "value");
			doReturn(itemMap).when(service).buildBescheidItemAsMap(any(), any());

			doUpdateBescheidDraft();

			verify(service).buildAttachedItemAsMap(command, itemMap);
		}

		@Test
		void shouldCallBuildUpdateAttachedItemCommand() {
			var expectedBodyObject = Map.<String, Object>of("key", "value");
			doReturn(expectedBodyObject).when(service).buildAttachedItemAsMap(any(), any());

			doUpdateBescheidDraft();

			verify(service).buildPatchAttachedItemCommand(bescheidItem, expectedBodyObject);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(updateAttachedItemCommand).when(service).buildPatchAttachedItemCommand(any(), any());

			doUpdateBescheidDraft();

			verify(commandService).createAndWaitUntilDone(updateAttachedItemCommand);
		}

		@Test
		void shouldThrowExceptionWhenNotFound() {
			when(remoteService.getItem(any())).thenThrow(StatusRuntimeException.class);

			assertThrows(StatusRuntimeException.class, this::doUpdateBescheidDraft);
		}

		private void doUpdateBescheidDraft() {
			service.updateBescheidDraft(command, document);
		}
	}

	@DisplayName("Build bescheid item")
	@Nested
	class TestBuildBescheidItemAsMap {

		private final Optional<Document> document = Optional.of(DocumentTestFactory.create());
		private final Command command = CommandTestFactory.createBuilder().bodyObject(AttachedItemTestFactory.createBescheidItem()).build();
		private final AttachedItem bescheidAttachedItem = AttachedItemTestFactory.createBescheid();

		@BeforeEach
		void mock() {
			doReturn(bescheidAttachedItem).when(service).getItem(any());
		}

		@Test
		void shouldBuildItemMapWithExistingBescheidFields() {
			service.buildBescheidItemAsMap(command, document);

			verify(service).buildItemMapWithExistingBescheidFields(command);
		}

		@Test
		void shouldGetItem() {
			service.buildBescheidItemAsMap(command, document);

			verify(service).getItem(CommandTestFactory.RELATION_ID);
		}

		@Test
		void shouldCallGetNewNachrichtText() {
			service.buildBescheidItemAsMap(command, document);

			verify(service).getNewNachrichtText(bescheidAttachedItem, document);
		}

		@Test
		void shouldOverrideIfNewNachrichtText() {
			doReturn(Optional.of(DocumentTestFactory.NACHRICHT_TEXT)).when(service).getNewNachrichtText(any(), any());

			var bescheidItem = service.buildBescheidItemAsMap(command, document);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_NACHRICHT_TEXT, DocumentTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldKeepNachrichtTextIfNotNew() {
			doReturn(Optional.empty()).when(service).getNewNachrichtText(any(), any());

			var bescheidItem = service.buildBescheidItemAsMap(command, document);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT);
		}
	}

	@DisplayName("Get new nachrichtText")
	@Nested
	class TestGetNewNachrichtText {

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();

		@DisplayName("with filled new document")
		@Nested
		class TestFilledDocument {

			private final Document document = DocumentTestFactory.create();
			private final Optional<Document> documentOpt = Optional.of(document);

			@Test
			void shouldCheckIfIsDifferentDocument() {
				service.getNewNachrichtText(bescheidItem, documentOpt);

				verify(service).isDifferentDocument(document, bescheidItem);
			}

			@DisplayName("and different documentId")
			@Nested
			class TestAndDifferentDocumentId {

				@BeforeEach
				void mock() {
					doReturn(true).when(service).isDifferentDocument(any(), any());
				}

				@Test
				void shouldReturnNachrichtTextFromDocument() {
					var nachrichtText = service.getNewNachrichtText(bescheidItem, documentOpt);

					assertThat(nachrichtText).hasValue(DocumentTestFactory.NACHRICHT_TEXT);
				}
			}

			@DisplayName("and same documentId")
			@Nested
			class TestAndSameDocumentId {

				@BeforeEach
				void mock() {
					doReturn(false).when(service).isDifferentDocument(any(), any());
				}

				@Test
				void shouldReturnNachrichtTextFromDocument() {
					var nachrichtText = service.getNewNachrichtText(bescheidItem, documentOpt);

					assertThat(nachrichtText).isEmpty();
				}
			}
		}

		@DisplayName("with empty document")
		@Nested
		class TestWithEmptyDocument {

			@Test
			void shouldReturnEmpty() {
				var nachrichtText = service.getNewNachrichtText(bescheidItem, Optional.empty());

				assertThat(nachrichtText).isEmpty();
			}
		}
	}

	@Nested
	class TestBuildAttachedItemAsMap {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

		@Test
		void shouldSetVorgangId() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(AttachedItem.PROPERTY_VORGANG_ID, command.getVorgangId());
		}

		@Test
		void shouldSetClient() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(AttachedItem.PROPERTY_CLIENT, BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		@Test
		void shouldSetItemName() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(AttachedItem.PROPERTY_ITEM_NAME, AttachedItemService.BESCHEID_ITEM_NAME);
		}

		@Test
		void shouldSetItem() {
			var expectedMap = Map.<String, Object>of("key", "value");

			var map = service.buildAttachedItemAsMap(command, expectedMap);

			assertThat(map).containsEntry(AttachedItem.PROPERTY_ITEM, expectedMap);
		}
	}

	@Nested
	class TestOverrideAttachedItem {

		private static final String CREATED_RESOURCE = "created-resource";

		@Mock
		private OzgCloudCommand updateItemCommand;

		@BeforeEach
		void init() {
			when(commandService.createAndWaitUntilDone(any())).thenReturn(updateItemCommand);
			when(updateItemCommand.getCreatedResource()).thenReturn(CREATED_RESOURCE);
		}

		@Test
		void shouldCallBuildUpdateAttachedItemCommand() {
			var bodyObject = Map.<String, Object>of("key", "value");
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem().item(bodyObject).build();

			service.overrideAttachedItem(bescheidItem);

			verify(service).buildPatchAttachedItemCommand(bescheidItem, bodyObject);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(updateItemCommand).when(service).buildPatchAttachedItemCommand(any(), any());

			service.overrideAttachedItem(AttachedItemTestFactory.createBescheid());

			verify(commandService).createAndWaitUntilDone(updateItemCommand);
		}

		@Test
		void shouldReturnCreatedResource() {
			var result = service.overrideAttachedItem(AttachedItemTestFactory.createBescheid());

			assertThat(result).isEqualTo(CREATED_RESOURCE);
		}
	}

	@DisplayName("Build map with all bescheid fields")
	@Nested
	class TestBuildMapWithAllBescheidFields {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

		@Test
		void shouldHaveSetBescheidenAm() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_BESCHIEDEN_AM, BescheidTestFactory.BESCHIEDEN_AM_STR);
		}

		@Test
		void shouldHaveSetBewilligt() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_BEWILLIGT, BescheidTestFactory.BEWILLIGT);
		}

		@Test
		void shouldHaveSetBescheidDocument() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_BESCHEID_DOCUMENT, DocumentTestFactory.ID);
		}

		@Test
		void shouldHaveSetSendBy() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_SEND_BY, BescheidTestFactory.SEND_BY.name());
		}

		@Test
		void shouldHaveSetNachrichtText() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldHaveSetNachrichtSubject() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldHaveSetAttachments() {
			var bescheidItem = service.buildItemMapWithAllBescheidFields(command);

			assertThat(bescheidItem).containsEntry(Bescheid.FIELD_ATTACHMENTS,
					List.of(BescheidTestFactory.ATTACHMENT_STR, BescheidTestFactory.ATTACHMENT_STR));
		}
	}

	@Nested
	class TestAddValueFromMapIfExists {

		private static final String KEY = "key";

		@Test
		void shouldAddValue() {
			var targetMap = new HashMap<String, Object>();
			var expectedValue = "value";
			Map<String, Object> sourceMap = Map.of(KEY, expectedValue);

			service.addValueFromMapIfExists(sourceMap, KEY, targetMap);

			assertThat(targetMap).containsEntry(KEY, expectedValue);
		}

		@Test
		void shouldNotAddValueIfKeyNotExists() {
			var targetMap = new HashMap<String, Object>();
			var expectedValue = "value";
			Map<String, Object> sourceMap = Map.of(KEY, expectedValue);

			service.addValueFromMapIfExists(sourceMap, "notExists", targetMap);

			assertThat(targetMap).isEmpty();
		}

		@Test
		void shouldAddNullValues() {
			var targetMap = new HashMap<String, Object>();
			Map<String, Object> sourceMap = Collections.singletonMap(KEY, null);

			service.addValueFromMapIfExists(sourceMap, KEY, targetMap);

			assertThat(targetMap).containsEntry(KEY, null);
		}
	}

	@Nested
	class TestBuildBescheidMap {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

		@Test
		void shouldSetStatus() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name());
		}

		@Test
		void shouldSetBeschiedenAm() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_BESCHIEDEN_AM, BescheidTestFactory.BESCHIEDEN_AM_STR);
		}

		@Test
		void shouldSetBewilligt() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_BEWILLIGT, true);
		}

		@Test
		void shouldSetBescheidDocument() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_BESCHEID_DOCUMENT, DocumentTestFactory.ID);
		}

		@Test
		void shouldSetSendBy() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_SEND_BY, BescheidTestFactory.SEND_BY.name());
		}

		@Test
		void shouldSetNachrichtText() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldSetAttachments() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(Bescheid.FIELD_ATTACHMENTS,
					List.of(BescheidTestFactory.ATTACHMENT_STR, BescheidTestFactory.ATTACHMENT_STR));
		}
	}

	@Nested
	class TestBuildUpdateAttachedItemCommand {

		@Test
		void shouldCallVorgangIdMapper() {
			service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangId() {
			var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
			when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

			var result = service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			assertThat(result.getVorgangId()).isEqualTo(expectedVorgangId);
		}

		@Test
		void shouldCallRelationIdMapper() {
			service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			verify(commandMapper).mapRelationId(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldSetRelationId() {
			var expectedId = GenericId.from("relationId");
			when(commandMapper.mapRelationId(any())).thenReturn(expectedId);

			var result = service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			assertThat(result.getRelationId()).isEqualTo(expectedId);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			assertThat(result.getRelationVersion()).isEqualTo(AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), Map.of());

			assertThat(result.getOrder()).isEqualTo(AttachedItemService.PATCH_ATTACHED_ITEM);
		}

		@Test
		void shouldSetBodyObject() {
			var bodyObject = Map.<String, Object>of("key", "value");

			var result = service.buildPatchAttachedItemCommand(AttachedItemTestFactory.createBescheid(), bodyObject);

			assertThat(result.getBodyObject()).containsAllEntriesOf(bodyObject);
		}

	}

	@Nested
	class TestDeleteBescheidDraft {

		@Mock
		private OzgCloudCommand deleteItemOzgCloudCommand;

		private final Command command = CommandTestFactory.createBuilder().relationId(AttachedItemTestFactory.ID).build();

		@Test
		void shouldCallGetBescheid() {
			when(remoteService.getItem(any())).thenReturn(AttachedItemTestFactory.createBescheid());

			deleteBescheidDraft();

			verify(remoteService).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(deleteItemOzgCloudCommand).when(service).buildDeleteItemCommand(any(), any());
			when(remoteService.getItem(any())).thenReturn(AttachedItemTestFactory.createBescheid());

			deleteBescheidDraft();

			verify(commandService).createAndWaitUntilDone(deleteItemOzgCloudCommand);
		}

		@Test
		void shouldThrowExceptionIfBescheidNotFound() {
			when(remoteService.getItem(any())).thenThrow(StatusRuntimeException.class);

			assertThrows(StatusRuntimeException.class, this::deleteBescheidDraft);
		}

		private void deleteBescheidDraft() {
			service.deleteBescheidDraft(command);
		}

		@Nested
		class TestBuildDeleteItemCommand {

			@Test
			void shouldCallCommandMapperWithVorgangId() {
				service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetVorgangId() {
				var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
				when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

				var ozgCloudCommand = service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				assertThat(ozgCloudCommand.getVorgangId()).isEqualTo(expectedVorgangId);
			}

			@Test
			void shouldCallCommandMapperWithRelationId() {
				service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				verify(commandMapper).mapRelationId(AttachedItemTestFactory.ID);
			}

			@Test
			void shouldSetRelationId() {
				var expectedId = GenericId.from(AttachedItemTestFactory.ID);
				when(commandMapper.mapRelationId(any())).thenReturn(expectedId);

				var ozgCloudCommand = service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				assertThat(ozgCloudCommand.getRelationId()).isEqualTo(expectedId);
			}

			@Test
			void shouldSetRelationVersion() {
				var ozgCloudCommand = service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				assertThat(ozgCloudCommand.getRelationVersion()).isEqualTo(AttachedItemTestFactory.VERSION);
			}

			@Test
			void shouldSetOrder() {
				var ozgCloudCommand = service.buildDeleteItemCommand(command, AttachedItemTestFactory.createBescheid());

				assertThat(ozgCloudCommand.getOrder()).isEqualTo(AttachedItemService.DELETE_ATTACHED_ITEM);
			}
		}
	}

	@Nested
	class TestValidateBescheid {

		@Test
		void shouldFilterDraftStatus() {
			assertDoesNotThrow(() -> service.validateBescheidStatus(AttachedItemTestFactory.createBescheid()));
		}

		@Test
		void shouldThrowExceptionIfBescheidIsNotDraft() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem().itemEntry(Bescheid.FIELD_STATUS, "otherStatus")
					.build();

			assertThrows(TechnicalException.class, () -> service.validateBescheidStatus(bescheidItem));
		}

	}

	@Nested
	class TestGetItem {

		@Test
		void shouldCallRemoteService() {
			service.getItem(AttachedItemTestFactory.ID);

			verify(service).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var expectedItem = AttachedItemTestFactory.createDocument();
			doReturn(expectedItem).when(remoteService).getItem(any());

			var result = service.getItem(AttachedItemTestFactory.ID);

			assertThat(result).isEqualTo(expectedItem);
		}
	}

	@Nested
	class TestBuildItemMap {

		@Test
		void shouldSetStatus() {
			var result = service.buildItemMap(Bescheid.Status.DRAFT);

			assertThat(result).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name());
		}
	}

	@Nested
	class TestFindAllBescheid {

		@Test
		void shouldCallRemoteService() {
			service.findAllBescheid(VorgangTestFactory.ID);

			verify(remoteService).findAllBescheid(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnResult() {
			var expectedBescheidItem = AttachedItemTestFactory.createBescheid();
			when(remoteService.findAllBescheid(any(VorgangId.class))).thenReturn(Collections.singletonList(expectedBescheidItem));

			var result = service.findAllBescheid(VorgangTestFactory.ID);

			assertThat(result).containsExactly(expectedBescheidItem);
		}
	}
}